from libifstate.util import logger
from libifstate.exception import ParserValidationError
from abc import ABC, abstractmethod
from copy import deepcopy


class Parser(ABC):
    _default_lo_link = {
        'name': 'lo',
        'addresses': [
            '127.0.0.1/8',
            '::1/128',
        ],
        'link': {
            'kind': 'physical',
            'state': 'up',
            'mtu': 65536,
        }
    }
    _default_ifstates = {
        'ignore': {
            'ipaddr_builtin': [
                'fe80::/10'
            ],
            'ipaddr_dynamic': True,
            'ifname_builtin': [
                r'^br-[\da-f]{12}',
                r'^docker\d+',
                r'^ppp\d+$',
                r'^veth',
                r'^virbr\d+',
                r'^vrrp\d*\.\d+$'
            ],
            'fdb_builtin': [
                r'^33:33:',
                r'^01:00:5e:'
            ],
            'routes_builtin': [
                {'proto': 1},
                {'proto': 2},
                {'proto': 8},
                {'proto': 9},
                {'proto': 10},
                {'proto': 11},
                {'proto': 12},
                {'proto': 13},
                {'proto': 14},
                {'proto': 15},
                {'proto': 16},
                {'proto': 18},
                {'proto': 42},
                {'proto': 186},
                {'proto': 187},
                {'proto': 188},
                {'proto': 189},
                {'proto': 192},
                {'to': 'ff00::/8'},
            ],
            'rules_builtin': [
                {'proto': 1},
                {'proto': 2},
                {'proto': 8},
                {'proto': 9},
                {'proto': 10},
                {'proto': 11},
                {'proto': 12},
                {'proto': 13},
                {'proto': 14},
                {'proto': 15},
                {'proto': 16},
                {'proto': 18},
                {'proto': 42},
                {'proto': 186},
                {'proto': 187},
                {'proto': 188},
                {'proto': 189},
                {'proto': 192},
            ],
        },
        'cshaper': {
            'default': {
                'egress_qdisc': {
                    'kind': 'cake',
                    'handle': '1:',
                },
                'ingress_qdisc': {
                    'kind': 'cake',
                    'handle': '1:',
                },
                'ingress_ifname': {
                    'search': r'^\D{1,3}',
                    'replace': 'ifb',
                }
            }
        }
    }

    @abstractmethod
    def __init__(self, name, **kwargs):
        self.ifstate = {}
        pass

    def merge(self, a, b):
        if not b is None:
            for key in b:
                if key in a:
                    if isinstance(a[key], dict) and isinstance(b[key], dict):
                        self.merge(a[key], b[key])
                    else:
                        a[key] = b[key]
                else:
                    a[key] = b[key]
        return a

    def _update_lo(self, cfg):
        if 'interfaces' in cfg:
            lo = next((idx for idx, iface in enumerate(cfg['interfaces']) if iface['name'] == 'lo'), None)
            if lo is not None:
                cfg['interfaces'][lo] = self.merge(deepcopy(Parser._default_lo_link), cfg['interfaces'][lo])
            else:
                cfg['interfaces'].append(Parser._default_lo_link)

    def config(self):
        # merge builtin defaults with config
        cfg = (self.merge(deepcopy(Parser._default_ifstates), self.ifstates))

        # 'ignore' should still be an object
        try:
            iter(cfg["ignore"])
        except TypeError:
            raise ParserValidationError("$.ignore: is not of type 'object'")

        # add loopback interface defaults
        self._update_lo(cfg)
        for namespace in cfg.get('namespaces', {}):
            self._update_lo(cfg['namespaces'][namespace])

        # merge builtin defaults
        for k in list(cfg["ignore"]):
            if k.endswith("_builtin"):
                n = k[:-8]
                if n in cfg["ignore"]:
                    try:
                        cfg["ignore"][n] += cfg["ignore"][k]
                    except TypeError:
                        raise ParserValidationError("$.ignore.{}: is not of type '{}'".format(
                            n, type(cfg["ignore"][k]).__name__))
                else:
                    cfg["ignore"][n] = cfg["ignore"][k]
                del(cfg["ignore"][k])

        return cfg



#
# The code in this file is mostly based on the Python package pybpfmaps by Peter Scholz
# which is distributed under the Apache 2.0 license.
#
#   https://github.com/PeterStolz/pybpfmaps
#

import ctypes
import os
import time
from enum import IntEnum
from libifstate.bpf.ctypes import libbpf

BPF_OBJ_NAME_LEN = 16


class Bpf_prog_info(ctypes.Structure):
    # info struct def https://github.com/torvalds/linux/blob/b4a5ea09b29371c2e6a10783faa3593428404343/tools/include/uapi/linux/bpf.h#L5880
    _fields_ = [
        ("type", ctypes.c_uint32),
        ("id", ctypes.c_uint32),
        ("tag", ctypes.c_uint8),
        ("jited_prog_len", ctypes.c_uint32),
        ("xlated_prog_len", ctypes.c_uint32),
        ("jited_prog_insns", ctypes.c_uint64),
        ("xlated_prog_insns", ctypes.c_uint64),
        ("load_time", ctypes.c_uint64),
        ("created_by_uid", ctypes.c_uint32),
        ("nr_map_ids", ctypes.c_uint32),
        ("map_ids", ctypes.c_uint64),
        ("name", ctypes.c_char * BPF_OBJ_NAME_LEN),
        ("ifindex", ctypes.c_uint32),
        ("gpl_compatible:1", ctypes.c_uint32),
        ("padding", ctypes.c_uint32),
        ("netns_dev", ctypes.c_uint64),
        ("netns_ino", ctypes.c_uint64),
        ("nr_jited_ksyms", ctypes.c_uint32),
        ("nr_jited_func_lens", ctypes.c_uint32),
        ("jited_ksyms", ctypes.c_uint64),
        ("jited_func_lens", ctypes.c_uint64),
        ("btf_id", ctypes.c_uint32),
        ("func_info_rec_size", ctypes.c_uint32),
        ("func_info", ctypes.c_uint64),
        ("nr_func_info", ctypes.c_uint32),
        ("nr_line_info", ctypes.c_uint32),
        ("line_info", ctypes.c_uint64),
        ("jited_line_info", ctypes.c_uint64),
        ("nr_jited_line_info", ctypes.c_uint32),
        ("line_info_rec_size", ctypes.c_uint32),
        ("jited_line_info_rec_size", ctypes.c_uint32),
        ("nr_prog_tags", ctypes.c_uint32),
        ("prog_tags", ctypes.c_uint64),
        ("run_time_ns", ctypes.c_uint64),
        ("run_cnt", ctypes.c_uint64),
        ("recursion_misses", ctypes.c_uint64),
        ("verified_insns", ctypes.c_uint32),
    ]


class Bpf_map_info(ctypes.Structure):
    # info struct def https://github.com/torvalds/linux/blob/b4a5ea09b29371c2e6a10783faa3593428404343/tools/include/uapi/linux/bpf.h#L5880
    _fields_ = [
        ("type", ctypes.c_uint32),
        ("id", ctypes.c_uint32),
        ("key_size", ctypes.c_uint32),
        ("value_size", ctypes.c_uint32),
        ("max_entries", ctypes.c_uint32),
        ("map_flags", ctypes.c_uint32),
        ("name", ctypes.c_char * BPF_OBJ_NAME_LEN),
        ("ifindex", ctypes.c_uint32),
        ("btf_vmlinux_value_type_id", ctypes.c_uint32),
        ("netns_dev", ctypes.c_uint64),
        ("netns_ino", ctypes.c_uint64),
        ("btf_id", ctypes.c_uint32),
        ("btf_key_type_id", ctypes.c_uint32),
        ("btf_value_type_id", ctypes.c_uint32),
        ("padding", ctypes.c_uint32),
        ("map_extra", ctypes.c_uint64),
    ]


class Bpf_map_struct(ctypes.Structure):
    """
    enum bpf_map_type map_type;
    __u32 key_size;
    __u32 value_size;
    __u32 max_entries;
    __u32 id;
    """

    _fields_ = [
        ("map_type", ctypes.c_uint32),
        ("key_size", ctypes.c_uint32),
        ("value_size", ctypes.c_uint32),
        ("max_entries", ctypes.c_uint32),
        ("id", ctypes.c_uint32),
    ]


class MapTypes(IntEnum):
    BPF_MAP_TYPE_HASH = 1
    BPF_MAP_TYPE_ARRAY = 2
    BPF_MAP_TYPE_PROG_ARRAY = 3
    BPF_MAP_TYPE_PERF_EVENT_ARRAY = 4
    BPF_MAP_TYPE_PERCPU_HASH = 5
    BPF_MAP_TYPE_PERCPU_ARRAY = 6
    BPF_MAP_TYPE_STACK_TRACE = 7
    BPF_MAP_TYPE_CGROUP_ARRAY = 8
    BPF_MAP_TYPE_LRU_HASH = 9
    BPF_MAP_TYPE_LRU_PERCPU_HASH = 10
    BPF_MAP_TYPE_LPM_TRIE = 11
    BPF_MAP_TYPE_ARRAY_OF_MAPS = 12
    BPF_MAP_TYPE_HASH_OF_MAPS = 13
    BPF_MAP_TYPE_DEVMAP = 14
    BPF_MAP_TYPE_SOCKMAP = 15
    BPF_MAP_TYPE_CPUMAP = 16
    BPF_MAP_TYPE_XSKMAP = 17
    BPF_MAP_TYPE_SOCKHASH = 18
    BPF_MAP_TYPE_CGROUP_STORAGE = 19
    BPF_MAP_TYPE_REUSEPORT_SOCKARRAY = 20
    BPF_MAP_TYPE_PERCPU_CGROUP_STORAGE = 21
    BPF_MAP_TYPE_QUEUE = 22
    BPF_MAP_TYPE_STACK = 23
    BPF_MAP_TYPE_SK_STORAGE = 24
    BPF_MAP_TYPE_DEVMAP_HASH = 25
    BPF_MAP_TYPE_STRUCT_OPS = 26
    BPF_MAP_TYPE_RINGBUF = 27
    BPF_MAP_TYPE_INODE_STORAGE = 28
    BPF_MAP_TYPE_TASK_STORAGE = 29

class KindTypes(IntEnum):
    BTF_KIND_INT = 1            # Integer
    BTF_KIND_PTR = 2            # Pointer
    BTF_KIND_ARRAY = 3          # Array
    BTF_KIND_STRUCT = 4         # Struct
    BTF_KIND_UNION = 5          # Union
    BTF_KIND_ENUM = 6           # Enumeration up to 32-bit values
    BTF_KIND_FWD = 7            # Forward
    BTF_KIND_TYPEDEF = 8        # Typedef
    BTF_KIND_VOLATILE = 9       # Volatile
    BTF_KIND_CONST = 10         # Const
    BTF_KIND_RESTRICT = 11      # Restrict
    BTF_KIND_FUNC = 12          # Function
    BTF_KIND_FUNC_PROTO = 13    # Function Proto
    BTF_KIND_VAR = 14           # Variable
    BTF_KIND_DATASEC = 15       # Section
    BTF_KIND_FLOAT = 16         # Floating point
    BTF_KIND_DECL_TAG = 17      # Decl Tag
    BTF_KIND_TYPE_TAG = 18      # Type Tag
    BTF_KIND_ENUM64 = 19        # Enumeration up to 64-bit values

class BPF_Map:
    def __init__(self, filename: str):
        self.__map_fd = libbpf.bpf_obj_get(os.fsencode(filename))
        if self.__map_fd < 0:
            raise OSError(-self.fd, "failed to access bpf map", filename)

        bpf_map_info = Bpf_map_info()
        rc = libbpf.bpf_obj_get_info_by_fd(self.fd,
                                           ctypes.byref(bpf_map_info),
                                           ctypes.byref(ctypes.c_uint(ctypes.sizeof(bpf_map_info))))
        assert(rc == 0)


        self.map_type = bpf_map_info.type
        self.map_name = bpf_map_info.name
        self.key_size = bpf_map_info.key_size
        self.value_size = bpf_map_info.value_size
        self.max_entries = bpf_map_info.max_entries
        self.map_flags = bpf_map_info.map_flags
        self.key_type = bpf_map_info.btf_key_type_id
        self.value_type = bpf_map_info.btf_value_type_id
        self.__id = bpf_map_info.id

    def __getitem__(self, key):
        """
        LIBBPF_API int bpf_map_lookup_elem(int fd, const void *key, void *value);
        """
        result = None
        if isinstance(key, slice):
            result = []
            for k in range(key.start, key.stop, key.step or 1):
                result.append(self[k])
        else:
            # Assume that if the provided key is not an int is is already a ctypes object
            if isinstance(key, int):
                key = ctypes.c_int(
                    key)# if self.key_type is None else self.key_type(key)
            value = ctypes.c_void_p(
                0)# if self.value_type is None else self.value_type()
            # import pudb; pudb.set_trace()
            err = libbpf.bpf_map_lookup_elem(
                ctypes.c_int(self.__map_fd), ctypes.byref(
                    key), ctypes.byref(value)
            )
            assert err == 0, f"Failed to lookup map elem {key}, {err}"
            if isinstance(value, ctypes.Structure):
                result = value
            else:
                result = value.value

            if result is None:
                result = 0
        return result

    def __setitem__(self, key, value):
        """
        LIBBPF_API int bpf_map_update_elem(int fd, const void *key, const void *value, __u64 flags);
        """
        # there is the option to specify a _as_parameter for custom classes, so they can be used when calling the function

        if isinstance(key, int):
            key = ctypes.c_int(
                key) if self.key_type is None else self.key_type(key)
        if isinstance(value, int):
            value = (
                ctypes.c_int(value)
                if self.value_type is None
                else self.value_type(value)
            )
        err = libbpf.bpf_map_update_elem(
            ctypes.c_int(self.__map_fd),
            ctypes.byref(key),
            ctypes.byref(value),
            ctypes.c_int(0),
        )
        assert err == 0, f"Failed to update map, {err}"

    def __len__(self):
        """Returns the number of max entries"""
        return self.max_entries

    def __iter__(self):
        if self.map_type == MapTypes.BPF_MAP_TYPE_HASH:
            return Bpf_map_iterator(self)
        elif self.map_type == MapTypes.BPF_MAP_TYPE_ARRAY:
            return Bpf_array_iterator(self)

    @property
    def id(self):
        return self.__id

    @property
    def fd(self):
        return self.__map_fd


class Bpf_array_iterator:
    def __init__(self, map: BPF_Map):
        self.__map = map
        self.__index = 0
        if map.map_type != MapTypes.BPF_MAP_TYPE_ARRAY:
            raise Exception("Only array maps are supported")

    def __iter__(self):
        return self

    def __next__(self):
        if self.__index >= len(self.__map):
            raise StopIteration
        else:
            result = self.__map[self.__index]
            self.__index += 1
            return result


class Bpf_map_iterator:
    def __init__(self, map: BPF_Map):
        self.__map = map
        if map.map_type != MapTypes.BPF_MAP_TYPE_HASH:
            raise Exception("Only hash maps are supported")
        # for the first iter we pass 0
        self.last_key = (
            ctypes.c_int(0)
        )
        self.first_iteration = True

    def __iter__(self):
        return self

    def __next__(self):
        # LIBBPF_API int bpf_map_get_next_key(int fd, const void *key, void *next_key);
        # strace og bpftool map dump
        # bpf(BPF_MAP_GET_NEXT_KEY, {map_fd=3, key=NULL, next_key=0x55fd70b6c2c0}, 128) = 0
        # bpf(BPF_MAP_LOOKUP_ELEM, {map_fd=3, key=0x55fd70b6c2c0, value=0x55fd70b6c2e0, flags=BPF_ANY}, 128) = 0
        # bpf(BPF_MAP_GET_NEXT_KEY, {map_fd=3, key=0x55fd70b6c2c0, next_key=0x55fd70b6c2c0}, 128) = 0
        # bpf(BPF_MAP_LOOKUP_ELEM, {map_fd=3, key=0x55fd70b6c2c0, value=0x55fd70b6c2e0, flags=BPF_ANY}, 128) = 0
        if self.first_iteration:
            self.next_exists = (
                libbpf.bpf_map_get_next_key(
                    self.__map.fd, 0, ctypes.byref(self.last_key)
                )
                == 0
            )
            self.first_iteration = False
        else:
            self.next_exists = (
                libbpf.bpf_map_get_next_key(
                    self.__map.fd,
                    ctypes.byref(self.last_key),
                    ctypes.byref(self.last_key),
                )
                == 0
            )
        if not self.next_exists:
            raise StopIteration
        value = self.__map[self.last_key]
        return self.last_key.value, value



import logging
from logging.handlers import QueueHandler, QueueListener
import os
import queue
import sys

logger = logging.getLogger('ifstate')
logger.propagate = False
logger.log_add = lambda option, oper='add': logger.info(oper, extra={'option': option, 'style': IfStateLogging.STYLE_CHG})
logger.log_change = lambda option, oper='change': logger.info(oper, extra={'option': option, 'style': IfStateLogging.STYLE_CHG})
logger.log_ok = lambda option, oper='ok': logger.info(oper, extra={'option': option, 'style': IfStateLogging.STYLE_OK})
logger.log_del = lambda option, oper='del': logger.info(oper, extra={'option': option, 'style': IfStateLogging.STYLE_DEL})

formatter = logging.Formatter('%(bol)s%(prefix)s%(style)s%(message)s%(eol)s')

class IfStateLogFilter(logging.Filter):
    def __init__(self, is_terminal):
        super().__init__()
        self.is_terminal = is_terminal

    def filter(self, record):
        record.levelshort = record.levelname[:1]

        if hasattr(record, 'iface'):
            record.prefix = " {:35} ".format(record.iface)
        else:
            record.prefix = ''

        if hasattr(record, 'option'):
            record.prefix = "   {:33} ".format(record.option)

        if self.is_terminal and record.levelno >= logging.WARNING:
            if record.levelno >= logging.ERROR:
                record.bol = IfStateLogging.ANSI_RED
            else:
                record.bol = IfStateLogging.ANSI_MAGENTA
            record.eol = IfStateLogging.ANSI_RESET
        else:
            record.bol = ''
            record.eol = ''

        if self.is_terminal and hasattr(record, 'style'):
            record.style = IfStateLogging.colorize(record.style)
            record.eol = IfStateLogging.ANSI_RESET
        else:
            record.style = ""

        return True


class IfStateLogging:
    STYLE_OK = "ok"
    STYLE_CHG = "chg"
    STYLE_DEL = "del"

    ANSI_GREEN = "\033[32m"
    ANSI_RED = "\033[31m"
    ANSI_MAGENTA = "\033[35m"
    ANSI_YELLOW = "\033[33m"
    ANSI_RESET = "\033[0m"
    ANSI_BOLD = "\033[1m"

    @staticmethod
    def colorize(style):
        if style == IfStateLogging.STYLE_OK:
            return IfStateLogging.ANSI_GREEN
        if style == IfStateLogging.STYLE_CHG:
            return IfStateLogging.ANSI_YELLOW
        if style == IfStateLogging.STYLE_DEL:
            return IfStateLogging.ANSI_YELLOW
        return ""

    def __init__(self, level, handlers=[], action=None):
        if level != logging.DEBUG:
            sys.tracebacklimit = 0

        logging.basicConfig(
            level=level,
        )

        has_stderr = sys.stderr is not None
        is_terminal = has_stderr and sys.stderr.isatty()

        # add custom logging handlers
        if not handlers:
            handlers = []

            if has_stderr:
                # log to stderr
                stream = logging.StreamHandler(sys.stderr)
                stream.addFilter(IfStateLogFilter(is_terminal))
                stream.setFormatter(formatter)
                handlers.append(stream)

            # log to syslog
            syslog = logging.handlers.SysLogHandler('/dev/log', facility=logging.handlers.SysLogHandler.LOG_DAEMON)
            if action is None:
                syslog.ident = 'ifstate[{}] '.format(os.getpid())
            else:
                syslog.ident = 'ifstate-{}[{}] '.format(action, os.getpid())
            syslog.addFilter(IfStateLogFilter(False))
            syslog.setFormatter(formatter)
            handlers.append(syslog)

        qu = queue.SimpleQueue()
        queue_handler = QueueHandler(qu)
        queue_handler.setLevel(level)
        logger.addHandler(queue_handler)
        self.listener = QueueListener(
            qu, *handlers, respect_handler_level=True)
        self.listener.start()

    def quit(self):
        self.listener.stop()



#!/usr/bin/env python3

from libifstate.parser import YamlParser
from libifstate import __version__, IfState
from libifstate.exception import FeatureMissingError, LinkNoConfigFound, LinkCircularLinked, ParserValidationError, ParserOpenError, ParserParseError, ParserIncludeError
from libifstate.util import logger, IfStateLogging
from collections import namedtuple
from copy import deepcopy
from setproctitle import setproctitle

import argparse
import logging
import re
import signal
import sys
import yaml


class Actions():
    CHECK = "check"
    APPLY = "apply"
    SHOW = "show"
    SHOWALL = "showall"
    VRRP = "vrrp"
    VRRP_FIFO = "vrrp-fifo"
    SHELL = "shell"

ACTIONS_HELP = {
    "CHECK"    : "dry run update the network config",
    "APPLY"    : "update the network config",
    "SHOW"     : "show running network config",
    "SHOWALL"  : "show running network config (more settings)",
    "VRRP"     : "run as keepalived notify script",
    "VRRP_FIFO": "run as keepalived notify_fifo_script",
    "SHELL"    : "launch interactive python shell (pyroute2)",
}

class IfsConfigHandler():
    def __init__(self, fn, soft_schema):
        self.fn = fn
        self.soft_schema = soft_schema

        self.ifs = self.load_config()

    def sighup_handler(self, signum, frame):
        logger.info("SIGHUP: reloading configuration")
        try:
            self.ifs = self.load_config()
        except Exception:
            pass

    def load_config(self):
        try:
            parser = YamlParser(self.fn)
        except ParserOpenError as ex:
            logger.error(
                "Config loading from {} failed: {}".format(ex.fn, ex.msg))
            raise ex
        except ParserParseError as ex:
            logger.error("Config parsing failed:\n\n{}".format(str(ex)))
            raise ex
        except ParserIncludeError as ex:
            logger.error(
                "Config include file {} failed: {}".format(ex.fn, ex.msg))
            raise ex

        try:
            ifstates = parser.config()

            ifs = IfState()
            ifs.update(ifstates, self.soft_schema)
            return ifs
        except ParserValidationError as ex:
            logger.error("Config validation failed for {}".format(ex.detail))
            raise ex
        except FeatureMissingError as ex:
            logger.error(
                "Config uses unavailable feature: {}".format(ex.feature))
            raise ex


def shell():
    from ifstate.shell import IfStateConsole
    import pyroute2

    shell = IfStateConsole()
    shell.interact(banner=f"ifstate {__version__}; pyroute2 {pyroute2.__version__}")


def main():
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group()
    parser.add_argument('--version', action='version',
                        version='%(prog)s {version}'.format(version=__version__))
    group.add_argument("-v", "--verbose", action="store_true",
                       help="be more verbose")
    group.add_argument("-q", "--quiet", action="store_true",
                       help="be more quiet, print only warnings and errors")
    parser.add_argument("-s", "--soft-schema", action="store_true",
                        help="ignore schema validation errors, expect ifstatecli to trigger internal exceptions")
    parser.add_argument("-c", "--config", type=str,
                        default="/etc/ifstate/config.yml", help="configuration YaML filename")
    subparsers = parser.add_subparsers(
        dest='action', required=True, help="specifies the action to perform")

    action_parsers = {
        a.lower().replace("_", "-"): subparsers.add_parser(a.lower().replace("_", "-"), help=ACTIONS_HELP[a]) for a in dir(Actions) if not a.startswith('_')
    }

    # Parameters for the vrrp action
    action_parsers[Actions.VRRP].add_argument(
        "type", type=str.lower, choices=["group", "instance"], help="type of vrrp notification")
    action_parsers[Actions.VRRP].add_argument(
        "name", type=str, help="name of the vrrp group or instance")
    action_parsers[Actions.VRRP].add_argument(
        "state", type=str.lower, choices=["unknown", "fault", "backup", "master"], help="the new state for the vrrp group or instance")

    # Parameters for the vrrp-fifo action
    action_parsers[Actions.VRRP_FIFO].add_argument(
        "fifo", type=str, help="named FIFO to read state changes from")

    args = parser.parse_args()
    if args.verbose:
        lvl = logging.DEBUG
    elif args.quiet:
        lvl = logging.ERROR
    else:
        lvl = logging.INFO

    if args.action == Actions.VRRP_FIFO:
        setproctitle("ifstate-{}@{}".format(args.action, args.fifo))
    else:
        setproctitle("ifstate-{}".format(args.action))

    if args.action == Actions.SHELL:
        shell()
        exit(0)

    ifslog = IfStateLogging(lvl, action=args.action)

    if args.action in [Actions.SHOW, Actions.SHOWALL]:
        # preserve dict order on python 3.7+
        if sys.version_info >= (3, 7):
            yaml.add_representer(
                dict, lambda self, data: yaml.representer.SafeRepresenter.represent_dict(self, data.items()))
        ifs = IfState()
        print(yaml.dump(ifs.show(args.action == Actions.SHOWALL)))

        ifslog.quit()
        exit(0)

    if args.action in [Actions.CHECK, Actions.APPLY, Actions.VRRP, Actions.VRRP_FIFO]:
        try:
            ifs_config = IfsConfigHandler(args.config, args.soft_schema)
        except (ParserOpenError,
                ParserParseError,
                ParserIncludeError,
                ParserValidationError,
                FeatureMissingError) as ex:
            ifslog.quit()
            exit(ex.exit_code())

        if args.action == Actions.CHECK:
            try:
                ifs_config.ifs.check()
            except LinkNoConfigFound:
                pass
            except LinkCircularLinked as ex:
                ifslog.quit()
                exit(ex.exit_code())
        elif args.action == Actions.VRRP_FIFO:
            signal.signal(signal.SIGHUP, ifs_config.sighup_handler)
            signal.signal(signal.SIGPIPE, signal.SIG_IGN)
            signal.signal(signal.SIGTERM, signal.SIG_IGN)

            status_pattern = re.compile(
                r'(group|instance) "([^"]+)" (unknown|fault|backup|master)( \d+)?$', re.IGNORECASE)

            with open(args.fifo) as fifo:
                for line in fifo:
                    m = status_pattern.match(line.strip())
                    if m:
                        try:
                            ifs_tmp = deepcopy(ifs_config.ifs)
                            ifs_tmp.apply(m.group(1), m.group(2), m.group(3))
                        except:
                            pass
        else:
            # ignore some well-known signals to prevent interruptions (i.e. due to ssh connection loss)
            signal.signal(signal.SIGHUP, signal.SIG_IGN)
            signal.signal(signal.SIGPIPE, signal.SIG_IGN)
            signal.signal(signal.SIGTERM, signal.SIG_IGN)
            try:
                if args.action == Actions.APPLY:
                    ifs_config.ifs.apply()
                elif args.action == Actions.VRRP:
                    ifs_config.ifs.apply(
                        args.type, args.name, args.state)
            except LinkNoConfigFound:
                pass
            except LinkCircularLinked as ex:
                ifslog.quit()
                exit(ex.exit_code())

        ifslog.quit()


if __name__ == "__main__":
    main()
